def get_schema():
    schema = {
        "setpoint-controller": {
            "description": "changes the set-point of the current program",
            "type": "object",
            "properties": {
                "topic-sub-down": {
                    "description": "reduce temperature topic",
                    "type": "string"
                },
                "command-down": {
                    "description": "down command - if this value is published to topic-sub-down, temp is reduced.",
                    "type": "string"
                },
                "topic-sub-up": {
                    "description": "increase temperature topic",
                    "type": "string"
                },
                "command-up": {
                    "description": "up command - if this value is published to topic-sub-up, temp is increased.",
                    "type": "string"
                },
                "topic-sub-reset": {
                    "description": "incoming event to reset temperature to default (optional together with "
                                   "command-reset)",
                    "type": "string"
                },
                "command-reset": {
                    "description": "command for topic-sub-reset / reset to default (optional together with "
                                   "topic-sub-reset)",
                    "type": "string"
                },
                "step-size": {
                    "description": "Temperature is changed by step size for each rotation step.",
                    "type": "number"
                },
                "max-temp": {
                    "description": "Maximum value for temperature",
                    "type": "number"
                },
                "min-temp": {
                    "description": "Minimum value for temperature",
                    "type": "number"
                },
            },
            "required": ["topic-sub-down", "command-down", "topic-sub-up", "command-up", "step-size",
                         "max-temp", "min-temp"],
            "additionalProperties": False
        }
    }

    return schema



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.controller.modes.modeprogram import ModeProgram
from argaeus.controller.modes.modeschedule import ModeSchedule
from argaeus.controller.modes.modefactory import ModeFactory
import collections


class Test_ModeFactory(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ModeFactory")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        mode_program_base = {
            "name": "",
            "type": "program",
            "selectable": True,
            "set-point": 19.5
        }

        topics_sub_config = {
            "to-left": "/test/rotate",
            "command-left": "LEFT",
            "to-right": "/test/rotate",
            "command-right": "RIGHT"
        }

        topics_pub_config = {
            "display-server-schedule-image": "/test/schedule",
            "display-server-mode-icon": "/test/modeicon",
            "temperature-set-point": "/test/setpoint"
        }

        img_config = {
            "width": 192,
            "height": 2,
            "foreground-color": 255,
            "background-color": 100,
            "patterns": {
                "A": 0,
                "B": 1,
                "C": 2,
                "D": 3
            }
        }

        mode_names = ["A", "B", "C", "D"]

        modes_config = []
        for name in mode_names:
            mode_program_base["name"] = name
            modes_config.append(mode_program_base.copy())

        modes_config[3]["selectable"] = False

        schedule_config = collections.OrderedDict()
        counter = 0
        for hour in range(24):
            for minute in range(0, 60, 15):
                name = "{:02d}:{:02d}".format(hour, minute)
                if counter < 96/4:
                    schedule_config[name] = mode_names[0]
                elif counter < 96/2:
                    schedule_config[name] = mode_names[1]
                elif counter < 96/4*3:
                    schedule_config[name] = mode_names[2]
                else:
                    schedule_config[name] = mode_names[3]
                counter += 1

        mode_schedule = {
            "name": "S",
            "type": "schedule",
            "selectable": True,
            "image": img_config,
            "schedule": schedule_config
        }
        modes_config.append(mode_schedule)

        self.config_base = {
            "default-mode": "S",
            "topics-sub": topics_sub_config,
            "topics-pub": topics_pub_config,
            "modes": modes_config
        }

    def test_0init(self):
        modes, modes_selectable = ModeFactory.create_modes(self.config_base["modes"], self.config_base["topics-pub"],
                                                           self.mqtt_client, self.logger)
        self.assertIsNotNone(modes)
        self.assertEqual(len(modes), 5)

        self.assertIsNotNone(modes_selectable)
        self.assertEqual(len(modes_selectable), 4)
        for mode in modes_selectable:
            self.assertNotEqual(mode.name, "D")
            self.assertNotEqual(mode.name, "d")

        self.assertListEqual(list(modes.keys()), ["A", "B", "C", "D", "S"])
        self.assertEqual(type(list(modes.values())[0]), ModeProgram)
        self.assertEqual(type(list(modes.values())[1]), ModeProgram)
        self.assertEqual(type(list(modes.values())[2]), ModeProgram)
        self.assertEqual(type(list(modes.values())[3]), ModeProgram)
        self.assertEqual(type(list(modes.values())[4]), ModeSchedule)
        self.assertEqual(type(modes["S"]), ModeSchedule)
        self.assertEqual(len(modes["S"].schedule_raw), 96)
        self.assertEqual(len(modes["S"].schedule), 96)
        self.assertIsNotNone(modes["S"].img)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
import threading
import time
from argaeus.controller.modes.modeprogram import ModeProgram


class Test_ModeProgram(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ModeProgram")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        self.mode_config = {
            "name": "Night",
            "type": "program",
            "selectable": True,
            "set-point": 19.5
        }

        self.topics_pub_config = {
            "display-server-schedule-image": "/test/schedule",
            "display-server-mode-icon": "/test/modeicon",
            "temperature-set-point": "/test/setpoint"
            }

    def test_0init(self):
        mp = ModeProgram(self.mode_config, self.topics_pub_config, self.mqtt_client, self.logger)
        self.assertIsNotNone(mp)
        self.assertEqual(mp.name, "Night")
        self.assertEqual(mp.selectable, True)
        self.assertEqual(mp.set_point, 19.5)
        self.assertEqual(mp.default_set_point, 19.5)
        self.assertEqual(mp._topic_pub_mode_icon, "/test/modeicon")
        self.assertEqual(mp._topic_pub_set_point, "/test/setpoint")

    def test_1mqtt(self):
        global modeicon_event
        modeicon_event = threading.Event()
        modeicon_event.clear()
        global setpoint_event
        setpoint_event = threading.Event()
        setpoint_event.clear()

        def setpoint(value):
            self.assertEqual(float(value), 19.5)
            setpoint_event.set()

        def modeicon(value):
            value = value.decode("UTF-8")
            self.assertEqual(value, "Night".lower())
            modeicon_event.set()

        self.mqtt_client.subscribe("/test/modeicon", modeicon)
        self.mqtt_client.subscribe("/test/setpoint", setpoint)

        mp = ModeProgram(self.mode_config, self.topics_pub_config, self.mqtt_client, self.logger)
        mp.publish()

        modeicon_event.wait(0.5)
        setpoint_event.wait(0.5)

        self.assertEqual(modeicon_event.is_set(), True)
        self.assertEqual(setpoint_event.is_set(), True)


if __name__ == '__main__':
    unittest.main()



from argaeus.controller.behavior.behaviorcontroller import BehaviorController
from argaeus.controller.busyledcontroller import BusyLedController
from argaeus.controller.modes.modecontroller import ModeController
from argaeus.controller.setpointcontroller import SetPointController
from argaeus.controller.operationcontroller import OperationController


class ControllerManager:
    _config = None
    _verbose = None

    _mode_controller = None
    _behavior_controller = None
    _busyled_controller = None
    _set_point_controller = None
    _operation_controller = None

    def __init__(self, config, verbose, state, mqtt_client):
        self._config = config
        self._verbose = verbose
        if self._verbose:
            print("ControllerManager.__init__ - initializing ControllerManager ('{}').".format(self._config))

        self._busyled_controller = BusyLedController(config["controller"]["busy-led-controller"], verbose, mqtt_client,
                                                     self._config["topics-sub"], self._config["topics-pub"],
                                                     self._config["mqtt-translations"])
        self._behavior_controller = BehaviorController(self._config["controller"]["button-controller"], self._verbose,
                                                       state, mqtt_client, self._config["topics-sub"],
                                                       self._config["mqtt-translations"])
        self._mode_controller = ModeController(self._config["controller"]["mode-controller"], self._verbose, state,
                                               mqtt_client, self._config["modes"], self._config["topics-sub"],
                                               self._config["topics-pub"], self._config["mqtt-translations"])
        self._set_point_controller = SetPointController(self._config["controller"]["setpoint-controller"],
                                                        self._verbose, state, mqtt_client, self._config["topics-sub"],
                                                        self._config["mqtt-translations"])
        self._operation_controller = OperationController(self._config["controller"]["operation-controller"],
                                                        self._verbose, state, mqtt_client, self._config["topics-pub"],
                                                        self._config["mqtt-translations"])
        if self._verbose:
            print("ControllerManager.__init__ - done.")

    def update(self):
        self._mode_controller.update()

    def start(self):
        self._busyled_controller.start()
        self._behavior_controller.start()
        self._mode_controller.start()
        self._set_point_controller.start()
        self._operation_controller.start()

    def stop(self):
        self._busyled_controller.stop()
        self._behavior_controller.stop()
        self._mode_controller.stop()
        self._set_point_controller.stop()
        self._operation_controller.stop()



def get_schema():
    schema = {
        "operation-controller": {
            "description": "turns the outgoing relais on/off",
            "type": "object",
            "properties": {
                "default-is-active": {
                    "description": "Is the controller active or passive initially",
                    "type": "boolean"
                },
                "topic-pub": {
                    "description": "Topic that controls the output behavior relais of the thermostat.",
                    "type": "string"
                },
                "command-active": {
                    "description": "set to active command - publish this value to topic-pub, to set the controller"
                                   "to active operation.",
                    "type": "string"
                },
                "command-passive": {
                    "description": "set to passive command - publish this value to topic-pub, to set the controller"
                                   "to passive operation.",
                    "type": "string"
                },
                "topic-sub-toggle": {
                    "description": "incoming event to toggle active/passive operation (optional together with "
                                   "command-toggle)",
                    "type": "string"
                },
                "command-toggle": {
                    "description": "command for topic-sub-toggle / toggle active/passive operation (optional "
                                   "together with topic-sub-toggle)",
                    "type": "string"
                }
            },
            "required": ["default-is-active", "topic-pub", "command-active", "command-passive"],
            "additionalProperties": False
        }
    }

    return schema



import argaeus.schema.setpointcontroller
import argaeus.schema.operationcontroller
import argaeus.schema.modecontroller


def add_property(properties, required, sub_schema):
    for key, value in sub_schema.items():
        properties[key] = value
        required.append(key)
    return properties, required


def get_schema():
    properties = {}
    required = []

    properties, required = add_property(properties, required, argaeus.schema.setpointcontroller.get_schema())
    properties, required = add_property(properties, required, argaeus.schema.operationcontroller.get_schema())
    properties, required = add_property(properties, required, argaeus.schema.modecontroller.get_schema())

    schema = {
        "controller": {
            "description": "Argaeus is the gui controller element for a thermostat.",
            "type": "object",
            "properties": properties,
            "required": required,
            "additionalProperties": False
        }
    }

    return schema



from pelops.mylogger import get_child


class AMode:
    """
    Abstract mode - a mode defines the current behavior of the thermostat.

    yaml config:
        name: Name  # name for this mode
        selectable: True  # True/False - if set to False it cannot be selected via GUI
        type: program  # ["program", "schedule"] - needed for mode factory.
    """

    _config = None  # config yaml structure
    _logger = None  # logger instance
    _mqtt_client = None  # mqtt_client instance

    name = None  # name of this mode
    selectable = None  # can this mode be activated via GUI

    def __init__(self, config, mqtt_client, logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mqtt_client instance
        :param logger: logger instance - a child will be spawned with the name defined in the config
        """

        self._config = config
        self._mqtt_client = mqtt_client
        self.name = self._config["name"]
        self._logger = get_child(logger, self.name)

        self._logger.info("{}.__init__ - initializing".format(self.name))
        self._logger.debug("{}.__init__ - config: '{}'.".format(self.name, self._config))

        self.selectable = bool(self._config["selectable"])

    def publish(self):
        """
        Publish your current state. abstract method.
        """
        raise NotImplementedError



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.controller.modes.modeprogram import ModeProgram
from argaeus.controller.modes.modeschedule import ModeSchedule
import collections
import datetime
import threading
from io import BytesIO
from PIL import Image


class Test_ModeSchedule(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ModeSchedule")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        mode_program_base = {
            "name": "",
            "type": "program",
            "selectable": True,
            "set-point": 19.5
        }

        self.topics_pub_config = {
            "display-server-schedule-image": "/test/schedule",
            "display-server-mode-icon": "/test/modeicon",
            "temperature-set-point": "/test/setpoint"
        }

        img_config = {
            "width": 192,
            "height": 2,
            "foreground-color": 255,
            "background-color": 100,
            "patterns": {
                "A": 0,
                "B": 1,
                "C": 2,
                "D": 3
            }
        }

        mode_names = ["A", "B", "C", "D"]

        self.program_modes = {}
        modes_config = []
        for name in mode_names:
            mode_program_base["name"] = name
            mp = ModeProgram(mode_program_base, self.topics_pub_config, self.mqtt_client, self.logger)
            self.program_modes[name] = mp
            modes_config.append(mode_program_base.copy())

        schedule_config = collections.OrderedDict()
        counter = 0
        for hour in range(24):
            for minute in range(0, 60, 15):
                name = "{:02d}:{:02d}".format(hour, minute)
                if counter < 96/4:
                    schedule_config[name] = self.program_modes["A"].name
                elif counter < 96/2:
                    schedule_config[name] = self.program_modes["B"].name
                elif counter < 96/4*3:
                    schedule_config[name] = self.program_modes["C"].name
                else:
                    schedule_config[name] = self.program_modes["D"].name
                counter += 1

        self.mode_schedule = {
            "name": "S",
            "type": "schedule",
            "selectable": True,
            "image": img_config,
            "schedule": schedule_config
        }

    def test_0init(self):
        ms = ModeSchedule(self.mode_schedule, self.topics_pub_config, self.mqtt_client, self.logger)
        self.assertIsNotNone(ms)
        self.assertEqual(ms._topic_pub, self.topics_pub_config["display-server-schedule-image"])
        self.assertEqual(len(ms.schedule), 0)
        self.assertEqual(len(ms.schedule_raw), 96)

        for k, value in ms.schedule_raw.items():
            key = "{:02d}:{:02d}".format(k.hour, k.minute)
            self.assertEqual(self.mode_schedule["schedule"][key], value)

        self.assertEqual(ms.selectable, True)
        self.assertEqual(ms.name, "S")
        self.assertIsNone(ms.img)

    def test_1sort_dict(self):
        sr_source = self.mode_schedule["schedule"].copy()
        self.assertEqual(len(sr_source), 96)

        # reverse order of entries
        sr_reverse = {}
        while len(sr_source) > 0:
            key, value = sr_source.popitem()
            sr_reverse[key] = value
        self.assertEqual(len(sr_reverse), 96)

        # check that order of entries is in fact reversed
        sr = self.mode_schedule["schedule"].copy()
        self.assertEqual(type(sr), collections.OrderedDict)
        self.assertEqual(type(sr_reverse), dict)
        self.assertNotEqual(list(sr.keys())[0], list(sr_reverse.keys())[0])
        self.assertNotEqual(list(sr.keys())[95], list(sr_reverse.keys())[95])
        self.assertEqual(list(sr.keys())[0], list(sr_reverse.keys())[95])
        self.assertEqual(list(sr.keys())[95], list(sr_reverse.keys())[0])

        # get ordered dict
        sr_target = ModeSchedule._sort_dict(sr_reverse)
        self.assertEqual(len(sr_target), 96)

        # check that order of entries is in fact ordered
        sr = self.mode_schedule["schedule"].copy()
        self.assertEqual(type(sr), collections.OrderedDict)
        self.assertEqual(type(sr_target), collections.OrderedDict)

        prev_key = None
        for key, value in sr_target.items():
            if prev_key is not None:
                self.assertTrue(prev_key<key)
            prev_key = key

    def test_2map(self):
        ms = ModeSchedule(self.mode_schedule, self.topics_pub_config, self.mqtt_client, self.logger)
        self.assertIsNotNone(ms)
        self.assertEqual(len(ms.schedule), 0)
        ms.map_schedule_modes(self.program_modes)
        self.assertEqual(len(ms.schedule), 96)
        self.assertIsNotNone(ms.img)

        for k,value in ms.schedule.items():
            key = "{:02d}:{:02d}".format(k.hour, k.minute)
            self.assertEqual(self.mode_schedule["schedule"][key], value.name)
            self.assertIn(value, self.program_modes.values())

    def test_3get_program_at_time(self):
        ms = ModeSchedule(self.mode_schedule, self.topics_pub_config, self.mqtt_client, self.logger)
        ms.map_schedule_modes(self.program_modes)

        counter = 0
        max_counter = 24*60
        for hour in range(24):
            for minute in range(60):
                key = datetime.time(hour=hour, minute=minute)
                program = ms.get_program_at_time(key)
                if counter < max_counter/4:
                    self.assertEqual(program, self.program_modes["A"])
                elif counter < max_counter/2:
                    self.assertEqual(program, self.program_modes["B"])
                elif counter < max_counter/4*3:
                    self.assertEqual(program, self.program_modes["C"])
                else:
                    self.assertEqual(program, self.program_modes["D"])
                counter += 1

    def test_4mqtt(self):
        on_message_event = threading.Event()
        on_message_event.clear()

        def _on_message(value):
            try:
                os.mkdir("modeschedule")
            except FileExistsError:
                pass

            mqtt_image = Image.open(BytesIO(value))
            mqtt_image = mqtt_image.convert("L")

            self.assertEqual(mqtt_image.size[0], 192)
            self.assertEqual(mqtt_image.size[1], 2)

            mqtt_image.save("modeschedule/modeschedule_0.png")
            on_message_event.set()

        self.mqtt_client.subscribe(self.topics_pub_config["display-server-schedule-image"], _on_message)

        ms = ModeSchedule(self.mode_schedule, self.topics_pub_config, self.mqtt_client, self.logger)
        ms.map_schedule_modes(self.program_modes)

        ms.publish()
        on_message_event.wait(0.5)
        self.assertTrue(on_message_event.is_set())


if __name__ == '__main__':
    unittest.main()



from pelops.mylogger import get_child
from PIL import Image
from PIL import ImageDraw


class ScheduleImage:
    """
    Generates a chart, that shows which mode is active at which time slot. It takes a list of 96 mode entries and a
    list of pattern mappings. For each mode entry, a pattern mapping must exist. The resulting image is stored in
    the property img.

    Four patterns exist:
        0 - nothing
        1 - lower dot
        2 - upper and lower dot
        3 - upper dot

    config yaml entries:
        width: 192  # width of image
        height: 2  # height of image
        foreground-color: 255  # from 0 to 255.
        background-color: 0  # from 0 to 255.
        patterns:  # 0, 1, 2, 3 are valid patterns; a pattern can be assigned more than once
            Night: 0    # nothing
            Morning: 1  # lower dot
            Day: 2      # upper and lower dot
            Frost: 3    # upper dot

    """

    _config = None  # yaml config structure
    _logger = None  # logger instance

    img = None  # image

    _width = None  # width of image
    _height = None  # height of image
    _foreground_color = None  # from 0 to 255
    _background_color = None  # from 0 to 255
    _group_by = None  # in minutes - defines the length of one time slot. currently fixed to 15.

    _pixel_per_slot = None  # x-axis
    _pixel_per_dot = None  # y-axis

    _pattern_schedule = None  # list with 96 entries where the modes have been replaced by their corresponding pattern

    def __init__(self, schedule_items, config, logger):
        """
        Constructor

        :param schedule_items: list of 96 mode entries
        :param config: config yaml structure
        :param logger: logger instance - a child with __name__ will be spawned
        """

        self._config = config
        self._logger = get_child(logger, __name__)

        self._logger.info("ScheduleImage - initializing")
        self._logger.debug("ScheduleImage - config: {}".format(self._config))
        self._logger.debug("ScheduleImage - schedule_items: {}".format(schedule_items))

        self._width = int(self._config["width"])
        self._height = int(self._config["height"])
        self._foreground_color = int(self._config["foreground-color"])
        self._background_color = int(self._config["background-color"])

        self.img = Image.new('L', (self._width, self._height), self._background_color)

        self._group_by = 15  # fixed step size for chart. -> 4 steps per hour / 24 hours -> 96 steps
        groups_per_hour = int(60 / self._group_by)
        pixel_per_hour = int(self._width / 24)
        self._pixel_per_slot = int(pixel_per_hour / groups_per_hour)

        if self._height % 2 != 0:
            self._logger.error("ScheduleImage - height ({}) must be dividable by 2.".format(self._height))
            raise ValueError("ScheduleImage - height ({}) must be dividable by 2.".format(self._height))
        self._pixel_per_dot = int(self._height / 2)

        self._set_pattern_schedule(schedule_items, self._config["patterns"])
        self._render_image()

    def _set_pattern_schedule(self, schedule_items, patterns):
        """
        Merge the 96 schedule items with the patterns from the config yaml structure to the pattern schedule needed for
        image generation.

        :param schedule_items: list of 96 mode entries
        :param patterns: dict with mode.entry as key and 0-3 as value
        """

        patterns = dict((k.lower(), v) for k, v in patterns.items())
        self._pattern_schedule = []

        for k,v in schedule_items.items():
            try:
                pattern = patterns[v.name.lower()]
                self._logger.info("ScheduleImage._set_pattern_schedule - [{}]: add pattern {}/{}.".
                                  format(k, v, pattern))
                self._pattern_schedule.append(pattern)
            except KeyError as e:
                self._logger.error("ScheduleImage._set_pattern_schedule - no pattern for schedule item '{}': {}".
                                   format(k, e))
                raise e

    def _render_image(self):
        """
        Generate image. Result is stored in property img.
        """

        self._logger.info("ScheduleImage._render_image() - pixel/dot: {}, pixel/slot: {}.".
                          format(self._pixel_per_dot, self._pixel_per_slot))
        # clear image
        draw = ImageDraw.Draw(self.img)
        draw.rectangle((0, 0, self._width, self._height), fill=self._background_color)
        # draw graph
        count = 0
        for pattern in self._pattern_schedule:
            self._logger.info("ScheduleImage._render_image - {}. pattern is {}.".format(count, pattern))
            x_from = count * self._pixel_per_slot
            x_to = x_from + self._pixel_per_slot - 1
            count = count + 1
            if pattern == 0: # nothing is displayed
                continue
            elif pattern == 1: # lower dot
                y_lower = self._height - 1
                y_upper = self._height - self._pixel_per_dot
            elif pattern == 2:  # upper dot
                y_lower = self._height - self._pixel_per_dot - 1
                y_upper = 0
            elif pattern == 3: # full height
                y_lower = self._height - 1
                y_upper = 0
            else:
                self._logger.error("ScheduleImage.update_image - unknown value for pattern '{}'.".format(pattern))
                raise ValueError("ScheduleImage.update_image - unknown value for pattern '{}'.".format(pattern))
            self._logger.info("ScheduleImage.update_image - rectangle({}/{}-{}/{}).".
                              format(count, pattern, x_from, y_lower, x_to, y_upper))
            draw.rectangle((x_from, y_lower, x_to, y_upper), fill=self._foreground_color)




import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.controller.operationcontroller import OperationController
import threading


class Test_OperationController(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_OperationController")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        pass

    def test_0init(self):
        oc = OperationController(self.config["controller"]["operation-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(oc)
        self.assertEqual(oc._default_is_active, oc.active_operation)
        self.assertEqual(oc._default_is_active, self.config["controller"]["operation-controller"]["default-is-active"])
        self.assertEqual(oc._topic_pub, self.config["controller"]["operation-controller"]["topic-pub"])
        self.assertEqual(oc._command_active, self.config["controller"]["operation-controller"]["command-active"])
        self.assertEqual(oc._command_passive, self.config["controller"]["operation-controller"]["command-passive"])
        self.assertEqual(oc._topic_sub_toggle, self.config["controller"]["operation-controller"]["topic-sub-toggle"])
        self.assertEqual(oc._command_sub_toggle, self.config["controller"]["operation-controller"]["command-toggle"])

    def test_1start_stop(self):
        ch = threading.Event()
        ch.clear()

        def _command_handler(value):
            value = value.decode("UTF-8")
            self.assertEqual(value, oc._command_active)
            ch.set()

        self.mqtt_client.subscribe(self.config["controller"]["operation-controller"]["topic-pub"], _command_handler)
        oc = OperationController(self.config["controller"]["operation-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(oc)

        oc.start()
        ch.wait()
        oc.stop()

    def test_2behavior(self):
        ch = threading.Event()
        ch.clear()

        states = []

        def _command_handler(value):
            value = value.decode("UTF-8")
            self.assertTrue(value == oc._command_active or value == oc._command_passive)
            states.append(value)
            ch.set()

        self.mqtt_client.subscribe(self.config["controller"]["operation-controller"]["topic-pub"], _command_handler)
        oc = OperationController(self.config["controller"]["operation-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(oc)

        oc.start()
        ch.wait()
        ch.clear()

        for i in range(4):
            self.mqtt_client.publish(self.config["controller"]["operation-controller"]["topic-sub-toggle"],
                                     self.config["controller"]["operation-controller"]["command-toggle"])
            ch.wait()
            ch.clear()

        oc.stop()
        self.assertListEqual(states, ['ON', 'OFF', 'ON', 'OFF', 'ON'])

    def test_3behavior_inverse(self):
        ch = threading.Event()
        ch.clear()

        states = []

        def _command_handler(value):
            value = value.decode("UTF-8")
            self.assertTrue(value == oc._command_active or value == oc._command_passive)
            states.append(value)
            ch.set()

        self.mqtt_client.subscribe(self.config["controller"]["operation-controller"]["topic-pub"], _command_handler)
        self.config["controller"]["operation-controller"]["default-is-active"] = False
        oc = OperationController(self.config["controller"]["operation-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(oc)

        oc.start()
        ch.wait()
        ch.clear()

        for i in range(4):
            self.mqtt_client.publish(self.config["controller"]["operation-controller"]["topic-sub-toggle"],
                                     self.config["controller"]["operation-controller"]["command-toggle"])
            ch.wait()
            ch.clear()

        oc.stop()
        self.assertListEqual(states, ['OFF', 'ON', 'OFF', 'ON', 'OFF'])


if __name__ == '__main__':
    unittest.main()



from enum import Enum


class ButtonBehavior(Enum):
    ToDefaultMode = 0  # change current mode to defined default mode
    ToggleActivePassive = 1  # change behavior of thermostat from observer to actuator (connect DAC to out)
    ResetToDefaultTemp = 2  # set-point for mode is changed to default temp

    @staticmethod
    def factory(name):
        if name.lower() == "to-default-mode":
            return ButtonBehavior.ToDefaultMode
        elif name.lower() == "toggle-active-passive":
            return ButtonBehavior.ToggleActivePassive
        elif name.lower() == "reset-to-default-temp":
            return ButtonBehavior.ResetToDefaultTemp
        else:
            raise ValueError("ButtonBehavior.factory - unknown value '{}'.".format(name))



from argaeus.controller.modes.modeprogram import ModeProgram
from argaeus.controller.modes.modeschedule import ModeSchedule
from pelops.mylogger import get_child


class ModeFactory:
    """
    Generation of mode instances
    """

    @staticmethod
    def create_mode(config, config_topics_pub, mqtt_client, logger):
        """
        Create a single mode instance

        :param config: config yaml structure
        :param config_topics_pub: a dict with all topics that can be published to
        :param mqtt_client: mqtt_client instance
        :param logger: logger instance
        :return: the corresponding mode instance
        """
        log = get_child(logger, __name__)
        log.info("ModeFactory.create_mode - creating mode ('{}').".format(config))

        t = config["type"].lower()
        if t == "program":
            mode = ModeProgram(config, config_topics_pub, mqtt_client, logger)
        elif t == "schedule":
            mode = ModeSchedule(config, config_topics_pub, mqtt_client, logger)
        else:
            log.error("ModeFactory.create_mode - unknown type '{}'.".format(t))
            raise ValueError("ModeFactory.create_mode - unknown type '{}'.".format(t))

        return mode

    @staticmethod
    def create_modes(config, config_topics_pub, mqtt_client, logger):
        """
        Takes a list of mode config entries and returns a list of mode instances. Schedules are already properly
        initialized (cp. two step initialization of ModeSchedule.

        :param config: config yaml structure
        :param config_topics_pub:  a dict with all topics that can be published to
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance
        :return: modes (dict with all generated instances - keys are mode.name), modes_selectable (a list with all all
        generated instances that can be selected via gui)
        """

        log = get_child(logger, __name__)
        log.info("ModeFactory.create_modes - creating modes ('{}').".format(config))

        modes = {}
        modes_selectable = []
        for c in config:
            mode = ModeFactory.create_mode(c, config_topics_pub, mqtt_client, logger)
            modes[mode.name] = mode
            if mode.selectable:
                modes_selectable.append(mode)
        log.info("ModeFactory.create_modes - map_schedule_modes")
        for mode in modes.values():
            if isinstance(mode, ModeSchedule):
                mode.map_schedule_modes(modes)

        return modes, modes_selectable



import datetime
from argaeus.controller.modes.modefactory import ModeFactory
from argaeus.controller.modes.modeschedule import ModeSchedule
from argaeus.controller.acontroller import AController


class ModeController(AController):
    """
    The core controller for the gui - it provides the current mode, program, and schedule:
      * current mode - either a program or a schedule. can be changed via GUI interaction
      * current program - if current mode is a program, then current program equals current mode. otherwise, the program
      that is currently active in the schedule defined by current mode is used as current program
      * current schedule - if current mode is of type ModeSchedule, current schedule is the raw schedule from current
      mode. 'None' oterwise.

    It registers to three topics:
      * prev - select the previous entry in the list of selectable modes as current mode
      * next - select the next entry in the list of selectable modes as current mode
      * default - select the default mode from config as current mode

    After each update, the current schedule chart (or None if no schedule is active) is published.

    config yaml entries:
        default-mode: Schedule  # default mode - must be a name from modes list
        topics-sub:  # incoming topics
            to-prev: /test/r2/rotate  # select previous mode
            command-prev: LEFT  # to previous command - if this value is published to to-prev, the previous entry in the mode list is selected
            to-next: /test/r2/rotate  # select next mode
            command-next: RIGHT  # to next command - if this value is published to to-next, the next entry in the mode list is selected
            to-default: /test/r1/button/pressed  # incoming event to reset to default mode (optional together with command-default)
            command-default: PRESSED  # command for topic-sub / reset to default mode (optional together with to-default)
        topics-pub:  # outgoing topics
            display-server-schedule-image: /test/display/schedule  # topic of an nikippe-mqttimage instance
            display-server-mode-icon: /test/display/mode  # topic of an nikippe-imagelist instance
            temperature-set-point: /test/temperature/set-point  # topic of e.g. epidaurus (=pid temperature control) set-point listener
        modes:  # list of modes
            - name: Night  # unique name for mode entry
              type: program  # program or schedule - a schedule consists of programms
              selectable: True  # can be selected using the gui
              set-point: 19.5  # target temperature of this mode

            - name: Day  # unique name for mode entry
              type: program  # program or schedule - a schedule consists of programms
              selectable: True  # can be selected using the gui
              set-point: 23.0  # target temperature of this mode

            - ...
    """

    _modes = None  # dict of modes with mode.name as key - result from ModeFactory
    _selectable_modes = None  # list of modes that can be selected via GUI
    _selectable_pos = None  # position of currently selected mode
    _default_mode = None  # default mode as specified in the config yaml

    _topic_sub_prev = None  # select previous mode
    _command_prev = None  # command expected on _topic_sub_prev
    _topic_sub_next = None  # select next mode
    _command_next = None  # command expected on _topic_sub_next
    _topic_sub_to_default = None  # reset to default mode - can be None
    _command_default = None  # command expected on _topic_sub_to_default - can be None

    _topic_pub_schedule_image = None  # publish current schedule chart to this topic

    current_mode = None  # current mode (program or schedule)
    current_program = None  # current program (either =current_mode or the current entry in schedule)
    current_schedule = None  # current schedule (either =current schedule or None)

    def __init__(self, config, mqtt_client, logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance - a child instance will be spawned with name=__name__
        """
        AController.__init__(self, config, mqtt_client, logger, logger_name=__name__)

        self._modes, self._selectable_modes = ModeFactory.create_modes(self._config["modes"],
                                                                       self._config["topics-pub"], mqtt_client, logger)

        self._default_mode = self._modes[self._config["default-mode"]]
        self._activate_default_mode()

        self._topic_sub_prev = self._config["topics-sub"]["to-prev"]
        self._command_prev = self._config["topics-sub"]["command-prev"]
        self._topic_sub_next = self._config["topics-sub"]["to-next"]
        self._command_next = self._config["topics-sub"]["command-next"]
        self._topic_pub_schedule_image = self._config["topics-pub"]["display-server-schedule-image"]

        try:
            self._topic_sub_to_default = self._config["topics-sub"]["to-default"]
        except KeyError:
            pass

        if self._topic_sub_to_default is not None:
            try:
                self._command_default = self._config["topics-sub"]["command-default"]
            except KeyError:
                self._logger.error("OperationController.__init__ - 'to-default' is set but 'command-default' "
                                   "is missing.")
                raise KeyError("OperationController.__init__ - 'to-default' is set but 'command-default' "
                               "is missing.")

        self._logger.info("ModeController.__init__ - done")

    def _activate_default_mode(self):
        """
        Set current mode/program/schedule to the default mode specified in config
        """

        self._logger.info("ModeController._activate_default_mode")
        self._selectable_pos = self._selectable_modes.index(self._default_mode)
        self.current_mode = self._default_mode

        if isinstance(self.current_mode, ModeSchedule):
            dt = datetime.datetime.time(datetime.datetime.now())
            self.current_program = self.current_mode.get_program_at_time(dt)
            self.current_schedule = self.current_mode.schedule_raw
        else:
            self.current_program = self.current_mode
            self.current_schedule = None

    def _to_default_handler(self, value):
        """
        Check if the incoming message on _topic_sub_to_default is equivalent to _topic_sub_to_default. Activate
        default mode if yes.

        :param value: mqtt message
        """
        if len(value) > 0 and value.decode("UTF-8") == self._command_default:
            self._logger.info("ModeController._to_default_handler - activate default mode")
            self._activate_default_mode()
            self.update()
        else:
            self._logger.warning("ModeController._to_default_handler - dont know how to handle "
                                 "message '{}'".format(value))

    def _topic_handler_prev(self, value):
        """
        Check if the incoming message on _topic_sub_prev is equivalent to _command_prev. Adapt selectable pos
        accordingly and call post topic handler.

        :param value: mqtt message
        """

        if value.decode("utf-8") == self._command_prev:
            self._logger.info("ModeController._topic_handler - command prev.")
            self._selectable_pos = self._selectable_pos - 1
            self._post_topic_handler()

    def _topic_handler_next(self, value):
        """
        Check if the incoming message on _topic_sub_next is equivalent to _command_next. Adapt selectable pos
        accordingly and call post topic handler.

        :param value: mqtt message
        """

        if value.decode("utf-8") == self._command_next:
            self._logger.info("ModeController._topic_handler - command next.")
            self._selectable_pos = self._selectable_pos + 1
            self._post_topic_handler()

    def _post_topic_handler(self):
        """
        Make sure that the selectable pos has a valid value and call update
        """
        self._selectable_pos = self._selectable_pos % len(self._selectable_modes)
        self.current_mode = self._selectable_modes[self._selectable_pos]
        self._logger.info("ModeController._topic_handler - selected mode '{}' at pos '{}'.".
                          format(self.current_mode.name, self._selectable_pos))
        self.update()

    def update(self):
        """
        fill current program and current schedule according to current mode and publish the program and the mode.
        if mode is not ModeSchedule publish an empty message to _topic_pub_schedule_image.
        """

        self._logger.info("ModeController.update")
        if isinstance(self.current_mode, ModeSchedule):
            dt = datetime.datetime.now().time()
            self.current_program = self.current_mode.get_program_at_time(dt)
            self.current_schedule = self.current_mode.schedule_raw
            self.current_program.publish()
            self.current_mode.publish()
        else:
            self.current_program = self.current_mode
            self.current_schedule = None
            self.current_program.publish()
            # no schedule active - publish empty message instead
            self._mqtt_client.publish(self._topic_pub_schedule_image, "")

    def start(self):
        """
        Subscribe to prev, next, and default
        """
        self._logger.info("ModeController.start - starting")
        self._mqtt_client.subscribe(self._topic_sub_prev, self._topic_handler_prev)
        self._mqtt_client.subscribe(self._topic_sub_next, self._topic_handler_next)
        if self._topic_sub_to_default is not None:
            self._mqtt_client.subscribe(self._topic_sub_to_default, self._to_default_handler)
        self.update()
        self._logger.info("ModeController.start - started")

    def stop(self):
        """
        Unsubscribe from prev, next, and default
        """
        self._logger.info("ModeController.start - stopping")
        self._mqtt_client.unsubscribe(self._topic_sub_prev, self._topic_handler_prev)
        self._mqtt_client.unsubscribe(self._topic_sub_next, self._topic_handler_next)
        if self._topic_sub_to_default is not None:
            self._mqtt_client.unsubscribe(self._topic_sub_to_default, self._to_default_handler)
        self._logger.info("ModeController.start - stopped")




from argaeus.controller.acontroller import AController


class SetPointController(AController):
    """
    The current set-point temperature can be changed with this controller. It subscribes to incoming events and increases
    /decreases the set-point of current_program from the mode_controller and initiates that the new value is published
    by the mode program itself. Optionally a button can be used to reset the current set-point to the default value
    specified in the config yaml.

    It registers to this topic:
      * down - reduces the set-point value
      * up - increases the set-point value
      * reset - reset to default value

    Each change is published to the pid controller.

    Config yaml entries:
        topic-sub-down: /test/r1/rotate  # reduce temperature topic
        command-down: LEFT  # down command - if this value is published to topic-sub-down, temp is reduced.
        topic-sub-up: /test/r1/rotate  # increase temperature topic
        command-up: RIGHT  # up command - if this value is published to topic-sub-up, temp is increased.
        topic-sub-reset: /test/r1/button/pressed  # incoming event to reset temperature to default (optional together with command-reset)
        command-reset: PRESSED  # command for topic-sub-reset / reset to default (optional together with topic-sub-reset)
        step-size: 0.5  # Temperature is changed by step size for each rotation step.
        max-temp: 30.0  # Maximum value for temperature
        min-temp: 10.0  # Minimum value for temperature
    """

    _step_size = None  # change of set_point per event
    _min_temp = None  # minimum value for set point
    _max_temp = None  # maximum value for set point

    _topic_sub_down = None  # topic for down event
    _command_down = None  # command for down event
    _topic_sub_up = None  # topic for up event
    _command_up = None  # command for up event

    _topic_sub_reset = None  # topic for reset to default event
    _command_reset = None  # command for reset to default event

    _mode_controller = None  # reference to mode controller - needed to get current mode

    def __init__(self, mode_controller, config, mqtt_client, logger):
        """
        Constructor

        :param mode_controller: reference to mode controller - needed to get current mode
        :param config: config yaml structure
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance - a child will be spawned with name=__name__
        """
        AController.__init__(self, config, mqtt_client, logger, logger_name=__name__)

        self._step_size = float(self._config["step-size"])
        self._min_temp = float(self._config["min-temp"])
        self._max_temp = float(self._config["max-temp"])

        self._topic_sub_down = self._config["topic-sub-down"]
        self._command_down = self._config["command-down"]
        self._topic_sub_up = self._config["topic-sub-up"]
        self._command_up = self._config["command-up"]

        try:
            self._topic_sub_reset = self._config["topic-sub-reset"]
        except KeyError:
            pass

        if self._topic_sub_reset is not None:
            try:
                self._command_reset = self._config["command-reset"]
            except KeyError:
                self._logger.error("OperationController.__init__ - 'topic-sub-reset' is set but 'command-reset' "
                                   "is missing.")
                raise KeyError("OperationController.__init__ - 'topic-sub-reset' is set but 'command-reset' is "
                               "missing.")

        self._mode_controller = mode_controller

        self._logger.info("SetPointController.__init__ - done")

    def _reset_temp_handler(self, value):
        """
        Check if the incoming message on _topic_sub_reset is equivalent to _topic_sub_reset. reset to default
        value if yes.

        :param value: mqtt message
        """
        if len(value) > 0 and value.decode("UTF-8") == self._command_reset:
            self._logger.info("SetPointController._reset_temp_handler - reset to default set point.")
            self._post_topic_handler(self._mode_controller.current_program.default_set_point)
        else:
            self._logger.warning("SetPointController._reset_temp_handler - dont know how to handle "
                                 "message '{}'".format(value))

    def _topic_handler_down(self, value):
        """
        if message is equal to command_down - reduce set point and call post processor
        :param value: mqtt message
        """
        if value.decode("utf-8") == self._command_down:
            self._logger.info("SetPointController._topic_handler - command left.")
            set_point = self._mode_controller.current_program.set_point - self._step_size
            self._post_topic_handler(set_point)

    def _topic_handler_up(self, value):
        """
        if message is equal to command_up - increase set point and call post processor
        :param value: mqtt message
        """
        if value.decode("utf-8") == self._command_up:
            self._logger.info("SetPointController._topic_handler - command right.")
            set_point = self._mode_controller.current_program.set_point + self._step_size
            self._post_topic_handler(set_point)

    def _post_topic_handler(self, set_to):
        """
        validate new set point and initialize publishing of the new value
        :param set_to:
        :return:
        """
        set_point = min(self._max_temp, max(self._min_temp, set_to))
        self._logger.info("SetPointController._topic_handler - set temp to '{}'.".
                          format(self._mode_controller.current_mode.name))
        self._mode_controller.current_program.set_point = set_point
        self._mode_controller.current_program.publish()

    def start(self):
        """start by subscribing to down, up, reset"""
        self._mqtt_client.subscribe(self._topic_sub_down, self._topic_handler_down)
        self._mqtt_client.subscribe(self._topic_sub_up, self._topic_handler_up)
        if self._topic_sub_reset is not None:
            self._mqtt_client.subscribe(self._topic_sub_reset, self._reset_temp_handler)

    def stop(self):
        """stop by unsubscribing from down, up, reset"""
        self._mqtt_client.unsubscribe(self._topic_sub_down, self._topic_handler_down)
        self._mqtt_client.unsubscribe(self._topic_sub_up, self._topic_handler_up)
        if self._topic_sub_reset is not None:
            self._mqtt_client.unsubscribe(self._topic_sub_reset, self._reset_temp_handler)



import collections
import datetime
from argaeus.controller.modes.amode import AMode
from argaeus.controller.modes.schedule_image import ScheduleImage
from io import BytesIO


class ModeSchedule(AMode):
    """
    A mode schedule provides for each time during a day the defined mode program. A day is divided into 15 minute
    slots. To each slot one of the mode programs is assigned. A schedule entry must not be another schedule. A
    schedule chart (see ScheduleImage) is generated that is published to a nikippe.mqttimage microservice for display
    purposes.

    Note: to create an instance of ModeSchedule, a two step initialisation is necessary: first call to constructor and
    as soon as all mode instances have been created, the method map_schedule_modes must be called. Calling
    get_program_at_time or publish before map_schedule_modes results in runtime errors.

    additional yaml config entries:
        image:
            ... # see schedule_image.py
        schedule:  # 24x4 entries that map the timeslot to a mode program
            "00:00": Night
            "00:15": Night
            "00:30": Night
            "00:45": Night
            ...
            "07:00": Morning
            "07:15": Morning
            "07:30": Morning
            "07:45": Morning
            ...
            "13:00": Day
            "13:15": Day
            "13:30": Day
            "13:45": Day
            ...
            "23:00": Night
            "23:15": Night
            "23:30": Night
            "23:45": Night
    """

    schedule_raw = None  # ordered dict with datetime.time as key and the string from the config as value
    schedule = None  # ordered dict with datetime.time as key and reference to the corresponding mode program as value
    img = None  # img containing the graphical representation of the schedule (see ScheduleImage)
    _mqtt_message = None  # img converted to bytes for mqtt transport
    _topic_pub = None  # publish schedule image to this topic
    _modes_mapped = None  # boolean - set to true as soon as the two step init phase has been finished

    def __init__(self, config, config_topics_pub, mqtt_client, logger):
        """
        Constructor

        :param config: config yaml structure
        :param config_topics_pub: a dict with all topics that can be published to
        :param mqtt_client: mqtt_client instance
        :param logger: logger instance
        """
        AMode.__init__(self, config, mqtt_client, logger)
        self.schedule_raw = ModeSchedule._sort_dict(self._config["schedule"])
        self.schedule = collections.OrderedDict()
        self._topic_pub = config_topics_pub["display-server-schedule-image"]
        self._modes_mapped = False

    @staticmethod
    def _sort_dict(d):
        """
        Takes a dict with "HH:MM" string as keys and any type as values and returns a time sorted ordered dict with
        datetime.time as keys.

        :param d: dict with "HH:MM" string keys and any value type
        :return: ordered dict with datetime.time as keys and any value type. ordered by time ascending.
        """
        result = collections.OrderedDict()
        for h in range (0, 24):
            for m in range(0, 60):
                key = "{:02}:{:02}".format(h, m)
                dt = datetime.time(hour=h, minute=m)
                try:
                    result[dt] = d[key]
                except KeyError:
                    pass
        return result

    def map_schedule_modes(self, modes):
        """
        Takes the raw schedule from the config, generates the final config with mode entry references as values and
        creates the schedule chart. Updates schedule, img and _modes_mapped.

        :param modes: dict with mode.name as key and mode instance as value
        """
        self._logger.info("ModeSchedule.map_schedule_modes - mapping schedule '{}' to mode '{}'.".
                          format(self.schedule_raw, modes))
        for k,v in self.schedule_raw.items():
            program = modes[v]
            if isinstance(program, ModeSchedule):
                self._logger.error("ModeSchedule.map_schedule_modes -  mode must not be of type ModeSchedule ('{}').".
                                   format(type(program)))
                raise ValueError("ModeSchedule.map_schedule_modes -  mode must not be of type ModeSchedule ('{}').".
                                 format(type(program)))
            self.schedule[k] = program

        self._logger.info("ModeSchedule.map_schedule_modes - generate schedule image")
        renderer = ScheduleImage(self.schedule, self._config["image"], self._logger)
        self.img = renderer.img
        bytes_image = BytesIO()
        self.img.save(bytes_image, format="png")
        self._mqtt_message = bytes_image.getvalue()

        self._modes_mapped = True

    def get_program_at_time(self, time):  #
        """
        Get the corresponding mode for any time during a day. Will return the same value for the same time on two
        different days.

        :param time: request time as datetime.time instance
        :return: mode instance from schedule
        """
        if not self._modes_mapped:
            self._logger.error("ModeSchedule.get_program_at_time - prior to calling this method, the method "
                               "'map_schedule_modes' must be called.")
            raise RuntimeError("ModeSchedule.get_program_at_time - prior to calling this method, the method "
                               "'map_schedule_modes' must be called.")

        result = None
        for dt, program in self.schedule.items():
            if dt > time:
                break
            result = program
        self._logger.info("ModeSchedule.get_program_at_time - program '{}' at time '{}'.".format(result.name, time))
        return result

    def publish(self):
        """
        Send the schedule image to a nikippe.mqttimage microservice.
        """
        self._logger.info("ModeSchedule.publish - sending schedule image")

        if not self._modes_mapped:
            self._logger.error("ModeSchedule.publish - prior to calling this method, the method "
                               "'map_schedule_modes' must be called.")
            raise RuntimeError("ModeSchedule.publish - prior to calling this method, the method "
                               "'map_schedule_modes' must be called.")

        self._logger.debug("ModeSchedule.publish - mqtt_message: {}".format(self._mqtt_message))
        self._mqtt_client.publish(self._topic_pub, self._mqtt_message)



from argaeus.controller.acontroller import AController


class OperationController(AController):
    """
    This controller is responsible whether the Themostat is operating in observation or in control mode. Thus, if the
    output of the DAC is connected to the heating system or not.

    It registers to this topic:
      * toggle - toggles the output between True and False

    Each change is published to the copreus controlled relay with the corresponding commands active and passive.

    Config yaml entries:
        default-is-active: True  # Is the controller active or passive initially
        topic-pub: /test/relais/closed  # Topic that controls the output behavior relais of the thermostat.
        command-active: ON  # set to active command - publish this value to topic-pub, to set the controller to
        active operation.
        command-passive: OFF  # set to passive command - publish this value to topic-pub, to set the controller to
        passive operation.
        topic-sub-toggle: /test/r1/button/pressed  # incoming event to toggle active/passive operation (optional
        together with command-toggle)
        command-toggle: PRESSED  # command for topic-sub-toggle / toggle active/passive operation (optional together
        with topic-sub-toggle)
    """
    _default_is_active = None  # default state as defined in the config
    active_operation = None  # current state
    
    _topic_pub = None  # publish state for relay to this topic
    _command_active = None  # activate DAC output
    _command_passive = None  # deactivate DAC output

    _topic_sub_toggle = None  # subscribe for incoming toggle trigger events
    _command_sub_toggle = None  # command expected to initiate trigger

    def __init__(self, config, mqtt_client, logger):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mymqttclient instance
        :param logger: logger instance - a child instance will be spawned with name=__name__
        """

        AController.__init__(self, config, mqtt_client, logger, logger_name=__name__)

        self._default_is_active = bool(self._config["default-is-active"])
        self.active_operation = self._default_is_active

        self._topic_pub = self._config["topic-pub"]
        self._command_active = self._config["command-active"]
        self._command_passive = self._config["command-passive"]

        try:
            self._topic_sub_toggle = self._config["topic-sub-toggle"]
        except KeyError:
            pass

        if self._topic_sub_toggle is not None:
            try:
                self._command_sub_toggle = self._config["command-toggle"]
            except KeyError:
                self._logger.error("OperationController.__init__ - 'topic-sub-toggle' is set but 'command-toggle' is "
                                   "missing.")
                raise KeyError("OperationController.__init__ - 'topic-sub-toggle' is set but 'command-toggle' "
                               "is missing.")

        self._logger.info("OperationController.__init__ - done")

    def _toggle_operation_handler(self, value):
        """
        Check if the incoming message on _topic_sub_toggle is equivalent to _command_sub_toggle. Toggle
        operation if yes.

        :param value: mqtt message
        """

        if len(value) > 0 and value.decode("UTF-8") == self._command_sub_toggle:
            self._toggle_operation()
        else:
            self._logger.warning("OperationController._toggle_operation_handler - dont know how to handle "
                                 "message '{}'".format(value))

    def _toggle_operation(self):
        """Change operation state to not previous state and initialize publish."""
        self.active_operation = not self.active_operation
        self._logger.info("OperationController._toggle_operation - toggle active/passive (now: '{}').".
                          format(self.active_operation))
        self._publish()

    def _publish(self):
        """
        publish current operation state (command_active or command_passive) to _topic_pub
        """
        if self.active_operation:
            self._mqtt_client.publish(self._topic_pub, self._command_active)
        else:
            self._mqtt_client.publish(self._topic_pub, self._command_passive)

    def start(self):
        """
        Subscribe topic_sub_toggle and publish initial state.
        """
        if self._topic_sub_toggle is not None:
            self._mqtt_client.subscribe(self._topic_sub_toggle, self._toggle_operation_handler)
        self._publish()

    def stop(self):
        """unsubcribe _topic_sub_toggle."""
        if self._topic_sub_toggle is not None:
            self._mqtt_client.unsubscribe(self._topic_sub_toggle, self._toggle_operation_handler)



from argaeus.controller.acontroller import AController


class BusyLedController(AController):
    _mqtt_client = None

    _topic_pub = None
    _topic_sub = None
    _command_on = None
    _command_off = None

    def __init__(self, config, verbose, mqtt_client, config_topics_sub, config_topics_pub, config_mqtt_translations):
        AController.__init__(self, config, verbose, config_topics_sub, config_topics_pub, config_mqtt_translations)
        self._mqtt_client = mqtt_client

        self._topic_sub = self._get_topic_sub(self._config["topic-sub"])
        self._topic_pub = self._get_topic_pub(self._config["topic-pub"])
        self._command_on = self._get_mqtt_translations(self._config["command-on"])
        self._command_off = self._get_mqtt_translations(self._config["command-off"])
        if self._verbose:
            print("BusyLedController.__init__ - done")

    def _topic_sub_handler(self, value):
        value = int(value)
        if value == 0:
            self._mqtt_client.publish(self._topic_pub, self._command_off)
        elif value > 0:
            self._mqtt_client.publish(self._topic_pub, self._command_on)
        else:
            raise ValueError("BusyLedController._topic_sub_handler - dont know what to do with value '{}'.".format(value))

    def start(self):
        self._mqtt_client.publish(self._topic_pub, self._command_off)
        self._mqtt_client.subscribe(self._topic_sub, self._topic_sub_handler)

    def stop(self):
        self._mqtt_client.unsubscribe(self._topic_sub, self._topic_sub_handler)



def get_schema():
    schema = {
        "description": "program entry",
        "type": "object",
        "properties": {
            "name": {
                "description": "unique name for mode entry",
                "type": "string"
            },
            "type": {
                "description": "type - must be 'program'",
                "type": "string",
                "enum": ["program"]
            },
            "selectable": {
                "description": "can be selected using the gui",
                "type": "boolean"
            },
            "set-point": {
                "description": "target temperature of this mode",
                "type": "number"
            }
        },
        "required": ["name", "type", "selectable", "set-point"],
        "additionalItems": False

    }

    return schema



from argaeus.controller.modes.modecontroller import ModeController
from argaeus.controller.setpointcontroller import SetPointController
from argaeus.controller.operationcontroller import OperationController
from pelops.abstractmicroservice import AbstractMicroservice
import time
from argaeus.schema.thermostatguicontroller import get_schema
import argaeus
import threading


class ThermostatGUIController(AbstractMicroservice):
    """
    Main component of argaeus - initializes the three controller: mode, set-point, and operation and updates the state
    every minute (e.g. check if the current schedule entry is a new mode entry).

    config yaml entries:
        setpoint-controller:
            ...
        mode-controller:
            ...
        operation-controller:
            ...
    """
    _version = argaeus.version  # version of software

    _loop_thread = None  # Thread instance containing _poll_loop

    _mode_controller = None  # mode controller instance
    _set_point_controller = None  # set point controller instance
    _operation_controller = None  # operation controller instance

    def __init__(self, config, mqtt_client=None, logger=None, stdout_log_level=None, no_gui=None):
        """
        Constructor

        :param config: config yaml structure
        :param mqtt_client: mymqttclient instance (optional)
        :param logger: logger instance (optionale)
        """
        AbstractMicroservice.__init__(self, config, "controller", mqtt_client=mqtt_client, logger=logger,
                                      logger_name=__name__, stdout_log_level=stdout_log_level, no_gui=no_gui)

        self._mode_controller = ModeController(self._config["mode-controller"], self._mqtt_client, self._logger)
        self._set_point_controller = SetPointController(self._mode_controller, self._config["setpoint-controller"],
                                                        self._mqtt_client, self._logger)
        self._operation_controller = OperationController(self._config["operation-controller"], self._mqtt_client,
                                                         self._logger)

        self._loop_thread = threading.Thread(target=self._poll_loop)

    @staticmethod
    def _calc_sleep_time():
        """
        calculate the seconds.milliseconds until the next full minute
        :return: seconds.milliseconds until next full minute
        """
        current_time = time.time()
        seconds_to_next_full_minute = (60 - current_time % 60)  # next full minute in seconds
        return seconds_to_next_full_minute

    def _poll_loop(self):
        """
        updates mode_controller each excatly each minute until _Stop_Service event.
        """
        self._logger.info("ThermostatGUIController._poll_loop - start")
        while not self._stop_service.isSet():
            self._mode_controller.update()
            sleep_for = ThermostatGUIController._calc_sleep_time()
            self._logger.info("ThermostatGUIController._poll_loop - sleeping for {} seconds.".format(sleep_for))
            self._stop_service.wait(timeout=sleep_for)
        self._logger.info("ThermostatGUIController._poll_loop - end")

    def _start(self):
        """starts the three controller and the loop"""
        self._mode_controller.start()
        self._set_point_controller.start()
        self._operation_controller.start()
        self._loop_thread.start()

    def _stop(self):
        """stops the loop and the three controller"""
        self._loop_thread.join()
        self._mode_controller.stop()
        self._set_point_controller.stop()
        self._operation_controller.stop()

    @classmethod
    def _get_schema(cls):
        return get_schema()

    @classmethod
    def _get_description(cls):
        return "Argaeus is the gui controller element for a thermostat."

    def runtime_information(self):
        return {}

    def config_information(self):
        return {}


def standalone():
    ThermostatGUIController.standalone()


if __name__ == "__main__":
    ThermostatGUIController.standalone()



from argaeus.controller.modes.amode import AMode


class ModeProgram(AMode):
    """
    A mode program sets the current temperature. It can either be selected via GUI or activated via a mode schedule.

    addtitional yaml config entry:
          set-point: 19.5  # target temperature of this mode
    """

    set_point = None  # current set_point for temperature
    default_set_point = None  # default set_point for temperature (=value taken from config)
    _topic_pub_set_point = None  # publish topic for set_point
    _topic_pub_mode_icon = None  # publish topic for mode_icon

    def __init__(self, config, config_topics_pub, mqtt_client, logger):
        """
        Constructor

        :param config: config yaml structure
        :param config_topics_pub: a dict with all topics that can be published to
        :param mqtt_client: mqtt_client instance
        :param logger: logger instance
        """

        AMode.__init__(self, config, mqtt_client, logger)
        self.set_point = float(self._config["set-point"])
        self.default_set_point = self.set_point
        self._topic_pub_set_point = config_topics_pub["temperature-set-point"]
        self._topic_pub_mode_icon = config_topics_pub["display-server-mode-icon"]

    def publish(self):
        """Publish set_point and name as mode_icon."""
        self._mqtt_client.publish(self._topic_pub_set_point, self.set_point)
        self._mqtt_client.publish(self._topic_pub_mode_icon, self.name.lower())




import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.myconfigtools import read_config, validate_config
from argaeus.thermostatguicontroller import ThermostatGUIController


class TestValidateConfig(unittest.TestCase):
    def test_validate_config(self):
        config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))+"/tests_unit/config.yaml")
        validation_result = validate_config(config, ThermostatGUIController.get_schema())
        self.assertIsNone(validation_result)


if __name__ == '__main__':
    unittest.main()



from argaeus.controller.behavior.buttonbehavior import ButtonBehavior


class Behavior:
    _config = None
    _verbose = None
    _mqtt_client = None
    _state = None

    _name = None
    _topic_sub = None
    _topic_command = None
    _behavior = None

    def __init__(self, config, verbose, state, mqtt_client, config_topics_sub, config_mqtt_translations):
        self._verbose = verbose
        self._config = config
        self._mqtt_client = mqtt_client
        self._state = state
        if self._verbose:
            print("Behavior.__init__ - initializing instance ('{}').".format(self._config))

        self._name = self._config["name"]

        group, element = self._config["topic-sub"].split(".")
        self._topic_sub = config_topics_sub[group][element]
        group, element = self._config["mqtt-translation"].split(".")
        self._topic_command = config_mqtt_translations[group][element]

        self._behavior = ButtonBehavior.factory(self._config["behavior"])

    def _topic_handler(self, value):
        if value.decode("utf-8")  == self._topic_command:
            if self._verbose:
                print("Behavior._topic_handler - detected behavior '{}' from input '{}'.".
                      format(self._behavior._name_, self._name))
            self._state.behavior_queues[self._behavior].put(True)
        else:
            if self._verbose:
                print("Behavior._topic_handler - received unknown value '{}'".format(value))

    def start(self):
        if self._verbose:
            print("Behavior.start - subscribing topic.")
        self._mqtt_client.subscribe(self._topic_sub, self._topic_handler)

    def stop(self):
        if self._verbose:
            print("Behavior.start - unsubscribing topic.")
        self._mqtt_client.unsubscribe(self._topic_sub, self._topic_handler)



def get_schema():
    schema = {
        "description": "schedule entry",
        "type": "object",
        "properties": {
            "name": {
                "description": "unique name for mode entry",
                "type": "string"
            },
            "type": {
                "description": "type - must be 'schedule'",
                "type": "string",
                "enum": ["schedule"]
            },
            "selectable": {
                "description": "can be selected using the rotary encoder",
                "type": "boolean"
            },
            "image": {
                "description": "generate image for nikippe",
                "type": "object",
                "properties": {
                    "width": {
                        "description": "width of image",
                        "type": "integer",
                        "minimum": 0
                    },
                    "height": {
                        "description": "height of image",
                        "type": "integer",
                        "minimum": 0
                    },
                    "foreground-color": {
                        "description": "foreground color",
                        "type": "integer",
                        "minimum": 0,
                        "maximum": 255
                    },
                    "background-color": {
                        "description": "background color",
                        "type": "integer",
                        "minimum": 0,
                        "maximum": 255
                    },
                    "patters": {
                        "description": "each property name must match one program-mode",
                        "type": "object",
                        "patternProperties": {
                            "^.*$": {
                                "description": "0 (draw nothing), 1 (lower dot), 2 (upper dot), 3 (both dots).",
                                "type": "integer",
                                "enum": [0, 1, 2, 3]
                            }
                        },
                    }
                },
                "required": ["width", "height", "foreground-color", "background-color", "patterns"],
                "additionalItems": False
            },
            "schedule": {
                "description": "definition which program is active in each 15 minute slot of a day",
                "type": "object",
                "properties": {},
                "required": [],
                "additionalItems": False
            },

        },
        "required": ["name", "type", "selectable", "image", "schedule"],
        "additionalItems": False
    }

    for hour in range(24):
        for minute in range(0, 60, 15):
            name = "{:02d}:{:02d}".format(hour, minute)
            schema["properties"]["schedule"]["properties"][name] = {
                "description": "Name of active program - must be taken from modes list.",
                "type": "string"
            }
            schema["properties"]["schedule"]["required"].append(name)

    return schema



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.thermostatguicontroller import ThermostatGUIController
import time
import threading


class Test_ThermostatGUIController(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ThermostatGUIController")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        self._wait()

    def _wait(self, minimum_seconds_to_next_minute=5):
        # guarantee that we will not have a racing condidtion like timestamp init is "00:14:59" and timestmap
        # assertion is "00:15:01" which could result in different schedule entries being used.
        wait = (60 - time.time() % 60) + 0.1
        if wait > minimum_seconds_to_next_minute:
            wait = 0  # everything is fine
        elif wait > 1:
            print("... waiting {} seconds before test starts.".format(wait))
        else:
            pass  # wait time is short enough - no need to bother user
        time.sleep(wait)

    def test_0init(self):
        tgc = ThermostatGUIController(self.config, self.mqtt_client, self.logger)
        self.assertIsNotNone(tgc)
        self.assertIsNotNone(tgc._operation_controller)
        self.assertIsNotNone(tgc._set_point_controller)
        self.assertIsNotNone(tgc._mode_controller)

    def test_1start_stop(self):
        self._wait(10)

        hsi = threading.Event()
        hmi = threading.Event()
        hsp = threading.Event()
        ch = threading.Event()
        hsi.clear()
        hmi.clear()
        hsp.clear()
        ch.clear()

        def _handler_schedule_image(value):
            hsi.set()

        def _handler_mode_icon(value):
            hmi.set()

        def _handler_set_point(value):
            hsp.set()

        def _command_handler(value):
            ch.set()

        self.mqtt_client.subscribe(self.config["controller"]["operation-controller"]["topic-pub"], _command_handler)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"],
            _handler_schedule_image)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-mode-icon"],
            _handler_mode_icon)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["temperature-set-point"],
            _handler_set_point)

        tgc = ThermostatGUIController(self.config, self.mqtt_client, self.logger, no_gui=True)
        tgc.start()

        hsi.wait()
        hmi.wait()
        hsp.wait()
        ch.wait()

        t = time.time()
        tgc.stop()
        diff = time.time() - t
        self.assertLess(diff, 2)  # stopping should be relatively fast - faster than the next poll intervall at least

    def test_1start_stop(self):
        self._wait(10)

        hsi = threading.Event()
        hmi = threading.Event()
        hsp = threading.Event()
        hsi.clear()
        hmi.clear()
        hsp.clear()

        def _handler_schedule_image(value):
            hsi.set()

        def _handler_mode_icon(value):
            hmi.set()

        def _handler_set_point(value):
            hsp.set()

        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"],
            _handler_schedule_image)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-mode-icon"],
            _handler_mode_icon)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["temperature-set-point"],
            _handler_set_point)

        tgc = ThermostatGUIController(self.config, self.mqtt_client, self.logger, no_gui=True)
        tgc.start()
        start_time = time.time()

        hsi.wait()
        hmi.wait()
        hsp.wait()
        hsi.clear()
        hmi.clear()
        hsp.clear()
        hsi.wait()
        hmi.wait()
        hsp.wait()
        hsi.clear()
        hmi.clear()
        hsp.clear()

        wait_time = (60 - start_time) % 60  # seconds to next update loop
        print("waiting for {} seconds for next event.".format(wait_time))
        hsi.wait(timeout=wait_time+0.1)  # add safety seconds
        hmi.wait()
        hsp.wait()
        self.assertTrue(hsi.is_set())
        self.assertTrue(hmi.is_set())
        self.assertTrue(hsp.is_set())

        self.assertLess(time.time(), start_time+wait_time+0.1)
        self.assertGreater(time.time(), start_time+wait_time-0.1)

        tgc.stop()


if __name__ == '__main__':
    unittest.main()




from argaeus.controller.acontroller import AController
from argaeus.controller.behavior.behavior import Behavior


class BehaviorController:
    _behaviors = None

    def __init__(self, config, verbose, state, mqtt_client, config_topics_sub, config_mqtt_translations):
        AController.__init__(self, config, verbose, config_topics_sub, None, config_mqtt_translations)

        self._behaviors = []
        for c in config:
            b = Behavior(c, verbose, state, mqtt_client, config_topics_sub, config_mqtt_translations)
            self._behaviors.append(b)

    def start(self):
        for b in self._behaviors:
            b.start()

    def stop(self):
        for b in self._behaviors:
            b.stop()



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config
from argaeus.controller.setpointcontroller import SetPointController
from argaeus.controller.modes.modecontroller import ModeController
import time, datetime
import math
import threading


class Test_OperationController(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_OperationController")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        self._wait()

    def _wait(self, minimum_seconds_to_next_minute=5):
        # guarantee that we will not have a racing condidtion like timestamp init is "00:14:59" and timestmap
        # assertion is "00:15:01" which could result in different schedule entries being used.
        wait = (60 - time.time() % 60) + 0.1
        if wait > minimum_seconds_to_next_minute:
            wait = 0  # everything is fine
        elif wait > 1:
            print("... waiting {} seconds before test starts.".format(wait))
        else:
            pass # wait time is short enough - no need to bother user
        time.sleep(wait)

    def test_0init(self):
        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        spc = SetPointController(mc, self.config["controller"]["setpoint-controller"], self.mqtt_client,
                                 self.logger)
        self.assertIsNotNone(spc)
        self.assertEqual(spc._step_size, float(self.config["controller"]["setpoint-controller"]["step-size"]))
        self.assertEqual(spc._min_temp, float(self.config["controller"]["setpoint-controller"]["min-temp"]))
        self.assertEqual(spc._max_temp, float(self.config["controller"]["setpoint-controller"]["max-temp"]))
        self.assertEqual(spc._topic_sub_down, self.config["controller"]["setpoint-controller"]["topic-sub-down"])
        self.assertEqual(spc._command_down, self.config["controller"]["setpoint-controller"]["command-down"])
        self.assertEqual(spc._topic_sub_up, self.config["controller"]["setpoint-controller"]["topic-sub-up"])
        self.assertEqual(spc._command_up, self.config["controller"]["setpoint-controller"]["command-up"])
        self.assertEqual(spc._topic_sub_reset, self.config["controller"]["setpoint-controller"]["topic-sub-reset"])
        self.assertEqual(spc._command_reset, self.config["controller"]["setpoint-controller"]["command-reset"])

    def test_1start_stop(self):
        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        spc = SetPointController(mc, self.config["controller"]["setpoint-controller"], self.mqtt_client,
                                 self.logger)
        self.assertIsNotNone(spc)
        spc.start()
        spc.stop()

    def test_2inc(self):
        sph = threading.Event()
        sph.clear()
        setpoints = []

        def _set_point_handler(value):
            setpoints.append(float(value))
            sph.set()

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        spc = SetPointController(mc, self.config["controller"]["setpoint-controller"], self.mqtt_client,
                                 self.logger)
        self.assertIsNotNone(spc)

        self.mqtt_client.subscribe(mc.current_program._topic_pub_set_point, _set_point_handler)

        spc.start()

        self.assertLess(mc.current_program.set_point, spc._max_temp-2)
        for i in range((math.ceil(spc._max_temp-mc.current_program.set_point)+1)*math.ceil(1/spc._step_size)):
            self.mqtt_client.publish(spc._topic_sub_up, spc._command_up)
            sph.wait()
            sph.clear()
        self.assertEqual(mc.current_program.set_point, spc._max_temp)

        while setpoints[len(setpoints)-2] == spc._max_temp:
            setpoints.pop()

        self.assertListEqual(setpoints[-7:], [27.0, 27.5, 28.0, 28.5, 29.0, 29.5, 30.0])
        spc.stop()

    def test_3dec(self):
        sph = threading.Event()
        sph.clear()
        setpoints = []

        def _set_point_handler(value):
            setpoints.append(float(value))
            sph.set()

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        spc = SetPointController(mc, self.config["controller"]["setpoint-controller"], self.mqtt_client,
                                 self.logger)
        self.assertIsNotNone(spc)

        self.mqtt_client.subscribe(mc.current_program._topic_pub_set_point, _set_point_handler)

        spc.start()

        self.assertGreater(mc.current_program.set_point, spc._min_temp+2)
        for i in range((math.ceil(mc.current_program.set_point-spc._min_temp)+1)*math.ceil(1/spc._step_size)):
            self.mqtt_client.publish(spc._topic_sub_down, spc._command_down)
            sph.wait()
            sph.clear()
        self.assertEqual(mc.current_program.set_point, spc._min_temp)

        while setpoints[len(setpoints)-2] == spc._min_temp:
            setpoints.pop()

        self.assertListEqual(setpoints[-7:], [13.0, 12.5, 12.0, 11.5, 11.0, 10.5, 10.0])
        spc.stop()

    def test_4reset(self):
        rh = threading.Event()
        rh.clear()

        def _reset_handler(value):
            rh.set()

        self.mqtt_client.subscribe(self.config["controller"]["setpoint-controller"]["topic-sub-reset"], _reset_handler)

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        spc = SetPointController(mc, self.config["controller"]["setpoint-controller"], self.mqtt_client,
                                 self.logger)
        self.assertIsNotNone(spc)
        spc.start()
        mc.start()
        mc.current_program.set_point = 10
        self.assertEqual(mc.current_program.set_point, 10)
        self.mqtt_client.publish(self.config["controller"]["setpoint-controller"]["topic-sub-reset"],
                                 self.config["controller"]["setpoint-controller"]["command-reset"])
        rh.wait()
        rh.clear()
        time.sleep(0.1)

        self.assertEqual(mc.current_program.set_point, mc.current_program.default_set_point)
        spc.stop()
        mc.stop()


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.controller.modes.modecontroller import ModeController
from PIL import Image, ImageChops
from io import BytesIO
import time, datetime
import math
import threading


class Test_ModeController(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ModeController")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        self._wait()

    def _wait(self, minimum_seconds_to_next_minute=5):
        # guarantee that we will not have a racing condidtion like timestamp init is "00:14:59" and timestmap
        # assertion is "00:15:01" which could result in different schedule entries being used.
        wait = (60 - time.time() % 60) + 0.1
        if wait > minimum_seconds_to_next_minute:
            wait = 0  # everything is fine
        elif wait > 1:
            print("... waiting {} seconds before test starts.".format(wait))
        else:
            pass # wait time is short enough - no need to bother user
        time.sleep(wait)

    @staticmethod
    def _equal(im1, im2):
        return ImageChops.difference(im1, im2).getbbox() is None

    def test_0init(self):
        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        self.assertEqual(len(mc._modes), 5)
        self.assertEqual(mc.current_mode.name, "Schedule")
        self.assertEqual(len(mc.current_schedule), 96)

        now = datetime.datetime.now().time()
        pos = math.floor(datetime.timedelta(hours=now.hour, minutes=now.minute)  / datetime.timedelta(minutes=15))
        current_program = list(mc.current_schedule.values())[pos]
        self.assertEqual(mc.current_program.name, current_program)

        self.assertEqual(mc._topic_sub_prev, self.config["controller"]["mode-controller"]["topics-sub"]["to-prev"])
        self.assertEqual(mc._command_prev, self.config["controller"]["mode-controller"]["topics-sub"]["command-prev"])
        self.assertEqual(mc._topic_sub_next, self.config["controller"]["mode-controller"]["topics-sub"]["to-next"])
        self.assertEqual(mc._command_next, self.config["controller"]["mode-controller"]["topics-sub"]["command-next"])
        self.assertEqual(mc._topic_pub_schedule_image,
                         self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"])
        self.assertEqual(mc._topic_sub_to_default,
                         self.config["controller"]["mode-controller"]["topics-sub"]["to-default"])
        self.assertEqual(mc._command_default,
                         self.config["controller"]["mode-controller"]["topics-sub"]["command-default"])

    def test_1start_stop(self):
        hsi = threading.Event()
        hsi.clear()
        hmi = threading.Event()
        hmi.clear()
        hsp = threading.Event()
        hsp.clear()

        def _handler_schedule_image(value):
            self.assertTrue(len(value)>0)
            mqtt_image = Image.open(BytesIO(value))
            mqtt_image = mqtt_image.convert("L")
            Test_ModeController._equal(mqtt_image, mc.current_mode.img)
            hsi.set()

        def _handler_mode_icon(value):
            self.assertEqual(value.decode("UTF-8"), mc.current_program.name.lower())
            hmi.set()

        def _handler_set_point(value):
            self.assertEqual(float(value), mc.current_program.set_point)
            hsp.set()

        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"],
            _handler_schedule_image)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-mode-icon"],
            _handler_mode_icon)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["temperature-set-point"],
            _handler_set_point)

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        mc.start()

        hsi.wait()
        hsp.wait()
        hmi.wait()

        mc.stop()

    def test_2prev_next(self):
        self._wait()

        def _clear_all():
            hmi.clear()
            hsi.clear()
            hsp.clear()

        def _wait_all():
            hmi.wait()
            hsi.wait()
            hsp.wait()

        hsi = threading.Event()
        hmi = threading.Event()
        hsp = threading.Event()

        _clear_all()

        schedule_images = []
        mode_icons = []
        set_points = []

        def _handler_schedule_image(value):
            if len(value)>0:
                mqtt_image = Image.open(BytesIO(value))
                mqtt_image = mqtt_image.convert("L")
                self.assertIsNotNone(mqtt_image)
                schedule_images.append(mqtt_image)
            else:
                schedule_images.append(None)
            hsi.set()

        def _handler_mode_icon(value):
            self.assertEqual(value.decode("UTF-8"), mc.current_program.name.lower())
            mode_icons.append(value.decode("UTF-8"))
            hmi.set()

        def _handler_set_point(value):
            self.assertEqual(float(value), mc.current_program.set_point)
            set_points.append(float(value))
            hsp.set()

        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"],
            _handler_schedule_image)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-mode-icon"],
            _handler_mode_icon)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["temperature-set-point"],
            _handler_set_point)

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        default_program = mc.current_program.name.lower()
        default_setpoint = mc.current_program.set_point
        mc.start()
        _wait_all()
        _clear_all()

        # turn counter-clockwise
        self.mqtt_client.publish(mc._topic_sub_prev, mc._command_prev)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_prev, mc._command_prev)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_prev, mc._command_prev)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_prev, mc._command_prev)
        _wait_all()
        _clear_all()

        # turn clockwise
        self.mqtt_client.publish(mc._topic_sub_next, mc._command_next)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_next, mc._command_next)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_next, mc._command_next)
        _wait_all()
        _clear_all()
        self.mqtt_client.publish(mc._topic_sub_next, mc._command_next)
        _wait_all()
        _clear_all()

        mc.stop()

        self.assertListEqual(set_points, [default_setpoint, 21.5, 23.0, 19.5, default_setpoint, 19.5, 23.0, 21.5,
                                          default_setpoint])
        self.assertListEqual(mode_icons, [default_program, 'morning', 'day', 'night', default_program, 'night', 'day',
                                          'morning', default_program])
        self.assertIsNotNone(schedule_images[0])
        self.assertIsNone(schedule_images[1])
        self.assertIsNone(schedule_images[2])
        self.assertIsNone(schedule_images[3])
        self.assertIsNotNone(schedule_images[4])
        self.assertIsNone(schedule_images[5])
        self.assertIsNone(schedule_images[6])
        self.assertIsNone(schedule_images[7])
        self.assertIsNotNone(schedule_images[8])

    def test_3to_default_mode(self):
        self._wait()

        def _clear_all():
            hmi.clear()
            hsi.clear()
            hsp.clear()

        def _wait_all():
            hmi.wait()
            hsi.wait()
            hsp.wait()

        hsi = threading.Event()
        hmi = threading.Event()
        hsp = threading.Event()

        _clear_all()

        def _handler_schedule_image(value):
            hsi.set()

        def _handler_mode_icon(value):
            hmi.set()

        def _handler_set_point(value):
            hsp.set()

        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-schedule-image"],
            _handler_schedule_image)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["display-server-mode-icon"],
            _handler_mode_icon)
        self.mqtt_client.subscribe(
            self.config["controller"]["mode-controller"]["topics-pub"]["temperature-set-point"],
            _handler_set_point)

        mc = ModeController(self.config["controller"]["mode-controller"], self.mqtt_client, self.logger)
        self.assertIsNotNone(mc)
        default_program = mc.current_program.name.lower()
        default_setpoint = mc.current_program.set_point

        mc.start()
        _wait_all()
        _clear_all()

        # turn counter-clockwise
        self.mqtt_client.publish(mc._topic_sub_prev, mc._command_prev)
        _wait_all()
        _clear_all()
        self.assertNotEqual(default_program, mc.current_program.name.lower())
        self.assertNotEqual(default_setpoint, mc.current_program.set_point)

        self.mqtt_client.publish(self.config["controller"]["mode-controller"]["topics-sub"]["to-default"],
                                 self.config["controller"]["mode-controller"]["topics-sub"]["command-default"])
        _wait_all()
        _clear_all()

        self.assertEqual(default_program, mc.current_program.name.lower())
        self.assertEqual(default_setpoint, mc.current_program.set_point)

        mc.stop()


if __name__ == '__main__':
    unittest.main()



from pelops.mylogger import get_child


class AController:
    """
    Abstract controller - provides the minimum functionality a controller must implement
    """

    _config = None  # config yaml structure
    _mqtt_client = None  # mqtt_client instance
    _logger = None  # logger instance

    def __init__(self, config, mqtt_client, logger, logger_name=__name__):
        """
        Constructor

        :param config: yaml structure
        :param mqtt_client: mqtt_client instance
        :param logger: logger instance - a child logger will be spawned
        :param logger_name: name for child logger
        """
        self._mqtt_client = mqtt_client
        self._config = config
        self._logger = get_child(logger, logger_name)

    def start(self):
        """Start operation - abstract method"""
        raise NotImplementedError

    def stop(self):
        """Stop operation - abstract method"""
        raise NotImplementedError



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
from pelops.myconfigtools import read_config, validate_config
from argaeus.controller.modes.schedule_image import ScheduleImage
from argaeus.controller.modes.modeprogram import ModeProgram
import collections


class Test_ScheduleImage(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = read_config(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) +
                                 "/tests_unit/config.yaml")
        cls.logger = create_logger(cls.config["logger"], "Test_ScheduleImage")
        cls.logger.info("start")
        cls.mqtt_client = MyMQTTClient(cls.config["mqtt"], cls.logger)
        cls.mqtt_client.connect()

    @classmethod
    def tearDownClass(cls):
        cls.mqtt_client.disconnect()
        cls.logger.info("end")

    def tearDown(self):
        self.mqtt_client.unsubscribe_all()

    def setUp(self):
        mode_config_base = {
            "name": "",
            "type": "program",
            "selectable": True,
            "set-point": 19.5
        }

        topics_pub_config = {
            "display-server-schedule-image": "/test/schedule",
            "display-server-mode-icon": "/test/modeicon",
            "temperature-set-point": "/test/setpoint"
        }

        mode_names = ["A", "B", "C", "D"]

        modes = {}
        for name in mode_names:
            mode_config_base["name"] = name
            mp = ModeProgram(mode_config_base, topics_pub_config, self.mqtt_client, self.logger)
            modes[name] = mp

        self.schedule_items = collections.OrderedDict()
        id = 0
        for name in mode_names:
            for i in range(int(96/4)):
                self.schedule_items[id] = modes[name]
                id += 1

        self.config_image = {
            "width": 192,
            "height": 2,
            "foreground-color": 255,
            "background-color": 100,
            "patterns": {
                "A": 0,
                "B": 1,
                "C": 2,
                "D": 3
            }
        }

    def _assertEqual_list(self, list, value):
        for entry in list:
            self.assertEqual(entry, value)

    def _get_horizontal_pixel(self, img, x, height):
        upper_pixel = []
        lower_pixel = []

        for y in range(height):
            if y < height/2:
                upper_pixel.append(img.getpixel((x, 0)))
            else:
                lower_pixel.append(img.getpixel((x, height - 1)))

        return lower_pixel, upper_pixel

    def _check_image(self, img, width, height):
        for x in range(width):
            lower_pixel, upper_pixel = self._get_horizontal_pixel(img, x, height)
            if x < width/2:
                self._assertEqual_list(upper_pixel, self.config_image["background-color"])
                if x < width/4:
                    self._assertEqual_list(lower_pixel, self.config_image["background-color"])
                else:
                    self._assertEqual_list(lower_pixel, self.config_image["foreground-color"])
            else:
                self._assertEqual_list(upper_pixel, self.config_image["foreground-color"])
                if x < width/4*3:
                    self._assertEqual_list(lower_pixel, self.config_image["background-color"])
                else:
                    self._assertEqual_list(lower_pixel, self.config_image["foreground-color"])

    def test_0image(self):
        width = self.config_image["width"]
        height = self.config_image["height"]
        si = ScheduleImage(self.schedule_items, self.config_image, self.logger)
        self.assertIsNotNone(si)
        self.assertIsNotNone(si.img)

        try:
            os.mkdir("scheduleimage")
        except FileExistsError:
            pass
        si.img.save("scheduleimage/scheduleimage_0.png")

        self.assertEqual(si.img.size[0], width)
        self.assertEqual(si.img.size[1], height)

        self._check_image(si.img, width, height)

    def test_1image_large(self):
        self.config_image["width"] = 192*2
        self.config_image["height"] = 12
        width = self.config_image["width"]
        height = self.config_image["height"]

        si = ScheduleImage(self.schedule_items, self.config_image, self.logger)
        self.assertIsNotNone(si)
        self.assertIsNotNone(si.img)

        try:
            os.mkdir("scheduleimage")
        except FileExistsError:
            pass
        si.img.save("scheduleimage/scheduleimage_1.png")

        self.assertEqual(si.img.size[0], width)
        self.assertEqual(si.img.size[1], height)

        self._check_image(si.img, width, height)

    def test_2not_enough_patterns(self):
        del self.config_image["patterns"]["D"]
        with self.assertRaises(KeyError):
            si = ScheduleImage(self.schedule_items, self.config_image, self.logger)


if __name__ == '__main__':
    unittest.main()



import argaeus.schema.modeprogram
import argaeus.schema.modeschedule


def get_schema():
    schema = {
        "mode-controller": {
            "description": "list of different operation modes like fixed temperature or time schedule driven",
            "type": "object",
            "properties": {
                "default-mode": {
                    "description": "default mode - must be a name from modes list",
                    "type": "string"
                },
                "topics-pub": {
                    "description": "outgoing topics",
                    "type": "object",
                    "properties": {
                        "display-server-schedule-image": {
                            "description": "topic of an nikippe-mqttimage instance",
                            "type": "string"
                        },
                        "display-server-mode-icon": {
                            "description": "topic of an nikippe-imagelist instance",
                            "type": "string"
                        },
                        "temperature-set-point": {
                            "description": "topic of e.g. epidaurus (=pid temperature control) set-point listener",
                            "type": "string"
                        },
                    },
                    "required": ["display-server-schedule-image", "display-server-mode-icon", "temperature-set-point"],
                    "additionalItems": False
                },
                "topics-sub": {
                    "description": "incoming topics",
                    "type": "object",
                    "properties": {
                        "to-prev": {
                            "description": "select previous mode",
                            "type": "string"
                        },
                        "command-prev": {
                            "description": "to previous command - if this value is published to to-prev, the previous "
                                           "entry in the mode list is selected",
                            "type": "string"
                        },
                        "to-next": {
                            "description": "select next mode",
                            "type": "string"
                        },
                        "command-next": {
                            "description": "to next command - if this value is published to to-next, the next "
                                           "entry in the mode list is selected",
                            "type": "string"
                        },
                        "to-default": {
                            "description": "incoming event to reset to default mode (optional together with "
                                           "command-default)",
                            "type": "string"
                        },
                        "command-default": {
                            "description": "command for topic-sub / reset to default mode (optional together with "
                                           "to-default)",
                            "type": "string"
                        }
                    },
                    "required": ["to-prev", "command-prev", "to-next", "command-next"],
                    "additionalItems": False
                },
                "modes": {
                    "description": "list of modes",
                    "type": "array",
                    "items": {
                        "anyof": [
                            argaeus.schema.modeschedule.get_schema(),
                            argaeus.schema.modeprogram.get_schema()
                        ]
                    }
                },

            },
            "required": ["default-mode", "topics-sub", "topics-pub", "modes"],
            "additionalProperties": False
        }
    }

    return schema



import threading


class BehaviorWorker:
    _state = None
    _verbose = None
    _EXECUTE_BEHAVIOR = None
    _behavior_executor = None
    _worker_thread = None

    def __init__(self, verbose, state, execute_behavior, behavior_executor):
        self._verbose = verbose
        self._state  = state
        self._EXECUTE_BEHAVIOR = execute_behavior
        self._behavior_executor = behavior_executor
        if self._verbose:
            print("BehaviorWorker.__init__ - initializing instance ('{}').".format(self._EXECUTE_BEHAVIOR))

        self._worker_thread = threading.Thread(target=self._worker)

    def _worker(self):
        if self._verbose:
            print("BehaviorWorker('{}') - started worker".format(self._EXECUTE_BEHAVIOR))
        while True:
            item = self._state.behavior_queues[self._EXECUTE_BEHAVIOR].get()
            if self._verbose:
                print("BehaviorWorker('{}') - worker received item '{}'.".format(self._EXECUTE_BEHAVIOR, item))
            if item is None:
                break
            self._behavior_executor(item)
            self._state.behavior_queues[self._EXECUTE_BEHAVIOR].task_done()
        if self._verbose:
            print("BehaviorWorker('{}') - stopped worker".format(self._EXECUTE_BEHAVIOR))

    def start(self):
        if self._verbose:
            print("BehaviorWorker.starting for '{}'".format(self._EXECUTE_BEHAVIOR))
        self._worker_thread.start()

    def stop(self):
        self._state.behavior_queues[self._EXECUTE_BEHAVIOR].put(None)
        self._worker_thread.join()
        if self._verbose:
            print("BehaviorWorker.stopped for '{}'".format(self._EXECUTE_BEHAVIOR))
